* Bit Manipulations, simulating the functionality of Clipper Tools functions
* using 100% Clipper.
* Optimized for execution speed.
* Dual 8-bit and 16-bit operation, 8-bit is much faster.

* Copyright 1997, Richard M. Mitchell, dba Bytemaster, All Rights Reserved

#include "common.ch"


*** FUNCTION TO BITWISE OR TWO NUMERIC BYTE OR 2-BYTE VALUES AND
*** RETURN THE RESULT
function NumOr( pnX, pnY, pl8Bit )
   local aX_Bits, ;
         aY_Bits, ;
         nZ := 0, ;
         i, ;
         aTwoPowered  := {1,2,4,8,16,32,64,128,256,512,1024,2048, ;
            4096,8192,16384,32768}, ;
         nBits

   default pl8Bit to .T.

   nBits := iif( pl8Bit, 8, 16 )

   aX_Bits := array( nBits )
   aY_Bits := array( nBits )

   *** LOOP FOR EACH BIT
   for i := nBits to 1 step -1

      *** PROCESS FIRST NUMBER
      if ( aX_Bits[i] := ;
            ! int( pnX / aTwoPowered[i] ) == 0 ) && ith BIT SET?
         pnX -= aTwoPowered[i]                   && REDUCE pnX
      endif

      *** PROCESS SECOND NUMBER
      if ( aY_Bits[i] := ;
            ! int( pnY / aTwoPowered[i] ) == 0 ) && ith BIT SET?
         pnY -= aTwoPowered[i]                   && REDUCE pnY
      endif

      *** CALCULATE BITWISE OR OF THE TWO BITS
      if aX_Bits[i] .OR. aY_Bits[i]
         nZ += aTwoPowered[i]                    && SET BIT
      endif

   next i

   return int( nZ )


*** FUNCTION TO BITWISE XOR TWO NUMERIC BYTE OR 2-BYTE VALUES AND
*** RETURN THE RESULT
function NumXOr( pnX, pnY, pl8Bit )
   local aX_Bits, ;
         aY_Bits, ;
         nZ := 0, ;
         i, ;
         aTwoPowered  := {1,2,4,8,16,32,64,128,256,512,1024,2048, ;
            4096,8192,16384,32768}, ;
         nBits

   default pl8Bit to .T.

   nBits := iif( pl8Bit, 8, 16 )

   aX_Bits := array( nBits )
   aY_Bits := array( nBits )

   *** LOOP FOR EACH BIT
   for i := nBits to 1 step -1

      *** PROCESS FIRST NUMBER
      if ( aX_Bits[i] := ;
            ! int( pnX / aTwoPowered[i] ) == 0 ) && ith BIT SET?
         pnX -= aTwoPowered[i]                   && REDUCE pnX
      endif

      *** PROCESS SECOND NUMBER
      if ( aY_Bits[i] := ;
            ! int( pnY / aTwoPowered[i] ) == 0 ) && ith BIT SET?
         pnY -= aTwoPowered[i]                   && REDUCE pnY
      endif

      *** CALCULATE BITWISE XOR OF THE TWO BITS
      if ( aX_Bits[i] .OR. aY_Bits[i] ) .AND. ;
            ! ( aX_Bits[i] .AND. aY_Bits[i] )
         nZ += aTwoPowered[i]                    && SET BIT
      endif

   next i

   return int( nZ )


*** FUNCTION TO BITWISE AND TWO NUMERIC BYTE OR 2-BYTE VALUES AND
*** RETURN THE RESULT
function NumAnd( pnX, pnY, pl8Bit )
   local aX_Bits, ;
         aY_Bits, ;
         nZ := 0, ;
         i, ;
         aTwoPowered  := {1,2,4,8,16,32,64,128,256,512,1024,2048, ;
            4096,8192,16384,32768}, ;
         nBits

   default pl8Bit to .T.

   nBits := iif( pl8Bit, 8, 16 )

   aX_Bits := array( nBits )
   aY_Bits := array( nBits )

   *** LOOP FOR EACH BIT
   for i := nBits to 1 step -1

      *** PROCESS FIRST NUMBER
      if ( aX_Bits[i] := ;
            ! int( pnX / aTwoPowered[i] ) == 0 ) && ith BIT SET?
         pnX -= aTwoPowered[i]                   && REDUCE pnX
      endif

      *** PROCESS SECOND NUMBER
      if ( aY_Bits[i] := ;
            ! int( pnY / aTwoPowered[i] ) == 0 ) && ith BIT SET?
         pnY -= aTwoPowered[i]                   && REDUCE pnY
      endif

      *** CALCULATE BITWISE XOR OF THE TWO BITS
      if ( aX_Bits[i] .AND. aY_Bits[i] )
         nZ += aTwoPowered[i]                    && SET BIT
      endif

   next i

   return int( nZ )


*** FUNCTION TO BITWISE NOT A NUMERIC BYTE OR 2-BYTE VALUE AND
*** RETURN THE RESULT
function NumNot( pnX, pl8Bit )

   default pl8Bit to .T.

   return iif( pl8Bit, 255 - pnX, 65535 - pnX )


* Note:  Bits are 7=128, 6=64, 5=32, 4=16, 3=8, 2=4, 1=2, 0=1
*   15=32768, 14=16384, 13=8192, 12=4096, 11=2048, 10=1024, 9=512, 8=256
function SetBit( pnNum, pnBit, pl8Bit )
   local nCalc := pnNum, ;
         aTwoPowered  := {1,2,4,8,16,32,64,128,256,512,1024,2048, ;
            4096,8192,16384,32768}, ;
         i, ;
         nBits

   default pl8Bit to .T.

   nBits := iif( pl8Bit, 8, 16 )

   *** LOOP FOR EACH BIT
   for i := nBits to ( pnBit + 1 ) step -1

      *** PROCESS FIRST NUMBER
      if ! int( pnNum / aTwoPowered[i] ) == 0      && ith BIT SET?
         pnNum -= aTwoPowered[i]                   && REDUCE pnNum
      elseif i == pnBit + 1
         nCalc += aTwoPowered[i]
      endif

   next i

   return nCalc


* Note:  Bits are 7=128, 6=64, 5=32, 4=16, 3=8, 2=4, 1=2, 0=1
*   15=32768, 14=16384, 13=8192, 12=4096, 11=2048, 10=1024, 9=512, 8=256
function ClearBit( pnNum, pnBit, pl8Bit )
   local nCalc := pnNum, ;
         aTwoPowered  := {1,2,4,8,16,32,64,128,256,512,1024,2048, ;
            4096,8192,16384,32768}, ;
         i, ;
         nBits

   default pl8Bit to .T.

   nBits := iif( pl8Bit, 8, 16 )

   *** LOOP FOR EACH BIT
   for i := nBits to ( pnBit + 1 ) step -1

      *** PROCESS FIRST NUMBER
      if ! int( pnNum / aTwoPowered[i] ) == 0      && ith BIT SET?
         pnNum -= aTwoPowered[i]                   && REDUCE pnNum
         if i == pnBit + 1
            nCalc -= aTwoPowered[i]
         endif
      endif

   next i

   return nCalc


function IsBit( pnNum, pnBit, pl8Bit )
   local nCalc := pnNum, ;
         aTwoPowered  := {1,2,4,8,16,32,64,128,256,512,1024,2048, ;
            4096,8192,16384,32768}, ;
         i, ;
         nBits, ;
         lCalc := .F.

   default pl8Bit to .T.

   nBits := iif( pl8Bit, 8, 16 )

   *** LOOP FOR EACH BIT
   for i := nBits to ( pnBit + 1 ) step -1

      *** PROCESS FIRST NUMBER
      if ! int( pnNum / aTwoPowered[i] ) == 0      && ith BIT SET?
         pnNum -= aTwoPowered[i]                   && REDUCE pnNum
         if i == pnBit + 1
            lCalc := .T.
         endif
      endif

   next i

   return lCalc


function LToN( plLogical )

   return iif( plLogical, 1, 0 )
